@@ -578,6 +578,8 @@ static int ntfs_filldir(ntfs_inode *dir_ni, s64 *pos, u8 ivcn_bits,
 		return 0;
 	if (ie->key.file_name.file_attributes & FILE_ATTR_I30_INDEX_PRESENT)
 		dt_type = NTFS_DT_DIR;
+	else if (fn->file_attributes & FILE_ATTR_SYSTEM)
+		dt_type = NTFS_DT_UNKNOWN;
 	else
 		dt_type = NTFS_DT_REG;
 	return filldir(dirent, fn->file_name, fn->file_name_length,
@@ -881,6 +883,7 @@ find_next_index_buffer:
 		if (bmp_buf_pos >> 3 < bmp_buf_size)
 			continue;
 		/* Read next chunk from the index bitmap. */
+		bmp_buf_pos = 0;
 		if ((bmp_pos >> 3) + bmp_buf_size > bmp_na->data_size)
 			bmp_buf_size = bmp_na->data_size - (bmp_pos >> 3);
 		br = ntfs_attr_pread(bmp_na, bmp_pos >> 3, bmp_buf_size, bmp);
@@ -1354,7 +1357,6 @@ no_hardlink:
 int ntfs_delete(ntfs_inode *ni, ntfs_inode *dir_ni, ntfschar *name, u8 name_len)
 {
 	ntfs_attr_search_ctx *actx = NULL;
-	ntfs_index_context *ictx = NULL;
 	FILE_NAME_ATTR *fn = NULL;
 	BOOL looking_for_dos_name = FALSE, looking_for_win32_name = FALSE;
 	BOOL case_sensitive_match = TRUE;
@@ -1449,19 +1451,7 @@ search:
 	if (ntfs_check_unlinkable_dir(ni, fn) < 0)
 		goto err_out;
 		
-	ictx = ntfs_index_ctx_get(dir_ni, NTFS_INDEX_I30, 4);
-	if (!ictx)
-		goto err_out;
-	if (ntfs_index_lookup(fn, le32_to_cpu(actx->attr->value_length), ictx))
-		goto err_out;
-	
-	if (((FILE_NAME_ATTR*)ictx->data)->file_attributes &
-			FILE_ATTR_REPARSE_POINT) {
-		errno = EOPNOTSUPP;
-		goto err_out;
-	}
-	
-	if (ntfs_index_rm(ictx))
+	if (ntfs_index_remove(dir_ni, fn, le32_to_cpu(actx->attr->value_length)))
 		goto err_out;
 	
 	if (ntfs_attr_record_rm(actx))
@@ -1483,8 +1473,10 @@ search:
 	 * case there are no reference to this inode left, so we should free all
 	 * non-resident attributes and mark all MFT record as not in use.
 	 */
-	if (ni->mrec->link_count)
-		goto out;
+	if (ni->mrec->link_count) {
+		ntfs_inode_update_times(ni, NTFS_UPDATE_CTIME);
+		goto ok;
+	}
 	ntfs_attr_reinit_search_ctx(actx);
 	while (!ntfs_attrs_walk(actx)) {
 		if (actx->attr->non_resident) {
@@ -1525,12 +1517,14 @@ search:
 				"Leaving inconsistent metadata.\n");
 	}
 	ni = NULL;
+ok:	
+	ntfs_inode_update_times(dir_ni, NTFS_UPDATE_MCTIME);
 out:
 	if (actx)
 		ntfs_attr_put_search_ctx(actx);
-	if (ictx)
-		ntfs_index_ctx_put(ictx);
-	if (ni && ntfs_inode_close(ni) && !err)
+	if (ntfs_inode_close(dir_ni) && !err)
+		err = errno;
+	if (ntfs_inode_close(ni) && !err)
 		err = errno;
 	if (err) {
 		errno = err;
@@ -1608,22 +1602,11 @@ int ntfs_link(ntfs_inode *ni, ntfs_inode *dir_ni, ntfschar *name, u8 name_len)
 	}
 	/* Add FILE_NAME attribute to inode. */
 	if (ntfs_attr_add(ni, AT_FILE_NAME, AT_UNNAMED, 0, (u8*)fn, fn_len)) {
-		ntfs_index_context *ictx;
-
-		err = errno;
 		ntfs_log_error("Failed to add FILE_NAME attribute.\n");
+		err = errno;
 		/* Try to remove just added attribute from index. */
-		ictx = ntfs_index_ctx_get(dir_ni, NTFS_INDEX_I30, 4);
-		if (!ictx)
+		if (ntfs_index_remove(dir_ni, fn, fn_len))
 			goto rollback_failed;
-		if (ntfs_index_lookup(fn, fn_len, ictx)) {
-			ntfs_index_ctx_put(ictx);
-			goto rollback_failed;
-		}
-		if (ntfs_index_rm(ictx)) {
-			ntfs_index_ctx_put(ictx);
-			goto rollback_failed;
-		}
 		goto err_out;
 	}
 	/* Increment hard links count. */
