@@ -409,7 +409,8 @@ ntfs_inode *ntfs_extent_inode_open(ntfs_inode *base_ni, const MFT_REF mref)
 					ni->mrec->sequence_number)) {
 				errno = EIO;
 				ntfs_log_perror("Found stale extent mft "
-					"reference mft=%lld", ni->mft_no);
+					"reference mft=%lld",
+					(long long)ni->mft_no);
 				return NULL;
 			}
 			return ni;
@@ -486,8 +487,7 @@ int ntfs_inode_attach_all_extents(ntfs_inode *ni)
 	while ((u8*)ale < ni->attr_list + ni->attr_list_size) {
 		if (ni->mft_no != MREF_LE(ale->mft_reference) &&
 				prev_attached != MREF_LE(ale->mft_reference)) {
-			if (!ntfs_extent_inode_open(ni,
-					MREF_LE(ale->mft_reference))) {
+			if (!ntfs_extent_inode_open(ni, ale->mft_reference)) {
 				ntfs_log_trace("Couldn't attach extent inode.\n");
 				return -1;
 			}
@@ -579,7 +579,7 @@ static int ntfs_inode_sync_file_name(ntfs_inode *ni)
 			if (!err)
 				err = errno;
 			ntfs_log_perror("Failed to open inode %lld with index",
-					le64_to_cpu(fn->parent_directory));
+				(long long)le64_to_cpu(fn->parent_directory));
 			continue;
 		}
 		ictx = ntfs_index_ctx_get(index_ni, NTFS_INDEX_I30, 4);
@@ -588,7 +588,8 @@ static int ntfs_inode_sync_file_name(ntfs_inode *ni)
 				err = errno;
 			ntfs_log_perror("Failed to get index ctx, inode %lld",
 					(long long)index_ni->mft_no);
-			ntfs_inode_close(index_ni);
+			if (ni != index_ni && ntfs_inode_close(index_ni) && !err)
+				err = errno;
 			continue;
 		}
 		if (ntfs_index_lookup(fn, sizeof(FILE_NAME_ATTR), ictx)) {
@@ -601,7 +602,8 @@ static int ntfs_inode_sync_file_name(ntfs_inode *ni)
 			ntfs_log_perror("Index lookup failed, inode %lld",
 					(long long)index_ni->mft_no);
 			ntfs_index_ctx_put(ictx);
-			ntfs_inode_close(index_ni);
+			if (ni != index_ni && ntfs_inode_close(index_ni) && !err)
+				err = errno;
 			continue;
 		}
 		/* Update flags and file size. */
@@ -686,7 +688,6 @@ int ntfs_inode_sync(ntfs_inode *ni)
 	/* Update FILE_NAME's in the index. */
 	if ((ni->mrec->flags & MFT_RECORD_IN_USE) && ni->nr_extents != -1 &&
 			NInoFileNameTestAndClearDirty(ni) &&
-			NInoParentMtimeUpdate(ni) &&
 			ntfs_inode_sync_file_name(ni)) {
 		if (!err || errno == EIO) {
 			err = errno;
@@ -1096,42 +1097,36 @@ put_err_out:
 }
 
 /**
- * ntfs_inode_update_atime - update access time for ntfs inode
- * @ni:		ntfs inode for which update access time
+ * ntfs_inode_update_times - update selected time fields for ntfs inode
+ * @ni:		ntfs inode for which update time fields
+ * @mask:	select which time fields should be updated
  *
- * This function usually get called when user read not metadata from inode.
- * Do not update time for system files.
+ * This function updates time fields to current time. Fields to update are
+ * selected using @mask (see enum @ntfs_time_update_flags for posssible values).
  */
-void ntfs_inode_update_atime(ntfs_inode *ni)
+void ntfs_inode_update_times(ntfs_inode *ni, ntfs_time_update_flags mask)
 {
-	if (!NVolReadOnly(ni->vol) && !NVolNoATime(ni->vol) && (ni->mft_no >=
-			FILE_first_user || ni->mft_no == FILE_root)) {
-		ni->last_access_time = time(NULL);
-		NInoFileNameSetDirty(ni);
-		NInoSetDirty(ni);
+	time_t now;
+
+	if (!ni) {
+		ntfs_log_error("%s(): Invalid arguments.\n", __FUNCTION__);
+		return;
 	}
-}
 
-/**
- * ntfs_inode_update_time - update all times for ntfs inode
- * @ni:		ntfs inode for which update times
- *
- * This function updates last access, mft and data change times. Usually
- * get called when user write not metadata to inode. Do not update time for
- * system files.
- */
-void ntfs_inode_update_time(ntfs_inode *ni)
-{
-	if (!NVolReadOnly(ni->vol) && NInoMtimeUpdate(ni) &&
-	    (ni->mft_no >= FILE_first_user || ni->mft_no == FILE_root)) {
-		time_t now;
+	if ((ni->mft_no < FILE_first_user && ni->mft_no != FILE_root) ||
+			NVolReadOnly(ni->vol) || !mask)
+		return;
 
-		now = time(NULL);
+	now = time(NULL);
+	if (mask & NTFS_UPDATE_ATIME)
+		ni->last_access_time = now;
+	if (mask & NTFS_UPDATE_MTIME)
 		ni->last_data_change_time = now;
+	if (mask & NTFS_UPDATE_CTIME)
 		ni->last_mft_change_time = now;
-		NInoFileNameSetDirty(ni);
-		NInoSetDirty(ni);
-	}
+	
+	NInoFileNameSetDirty(ni);
+	NInoSetDirty(ni);
 }
 
 /**
