@@ -4,7 +4,7 @@
  * Copyright (c) 2002-2005 Richard Russon
  * Copyright (c) 2003-2006 Anton Altaparmakov
  * Copyright (c) 2003 Lode Leroy
- * Copyright (c) 2005-2007 Szabolcs Szakacsits
+ * Copyright (c) 2005-2008 Szabolcs Szakacsits
  *
  * A set of shared functions for ntfs utilities
  *
@@ -62,7 +62,12 @@ static const char *corrupt_volume_msg =
 
 static const char *hibernated_volume_msg =
 "The NTFS partition is hibernated. Please resume and shutdown Windows\n"
-"properly, so mounting could be done safely.\n";
+"properly, or mount the volume read-only with the 'ro' mount option, or\n"
+"mount the volume read-write with the 'remove_hiberfile' mount option.\n"
+"For example type on the command line:\n"
+"\n"
+"            mount -t ntfs-3g %s %s -o remove_hiberfile\n"
+"\n";
 
 static const char *unclean_journal_msg =
 "Mount is denied because NTFS is marked to be in use. Choose one action:\n"
@@ -84,13 +89,18 @@ static const char *fakeraid_msg =
 "different device under /dev/mapper/, (e.g. /dev/mapper/nvidia_eahaabcc1)\n"
 "to mount NTFS. Please see the 'dmraid' documentation for help.\n";
 
+static const char *access_denied_msg =
+"Please check the volume and the ntfs-3g binary permissions,\n"
+"and the mounting user ID. More explanation is provided at\n"
+"http://ntfs-3g.org/support.html#unprivileged\n";
+
 static const char *forced_mount_msg =
 "\n"
 "            mount -t ntfs-3g %s %s -o force\n"
 "\n"
 "    Or add the option to the relevant row in the /etc/fstab file:\n"
 "\n"
-"            %s %s ntfs-3g defaults,force 0 0\n";
+"            %s %s ntfs-3g force 0 0\n";
 
 /**
  * utils_set_locale
@@ -109,34 +119,32 @@ int utils_set_locale(void)
 	return 0;
 }
 
-ntfs_volume *utils_mount_volume(const char *volume, const char *mntpoint,
-				unsigned long flags)
+void utils_mount_error(const char *volume, const char *mntpoint, int err)
 {
-	ntfs_volume *vol;
-
-	vol = ntfs_mount(volume, flags);
-	if (!vol) {
-		
-		ntfs_log_perror("Failed to mount '%s'", volume);
-		
-		if (errno == EINVAL)
+	switch (err) {
+		case NTFS_VOLUME_NOT_NTFS:
 			ntfs_log_error(invalid_ntfs_msg, volume);
-		else if (errno == EIO)
+			break;
+		case NTFS_VOLUME_CORRUPT:
 			ntfs_log_error("%s", corrupt_volume_msg);
-		else if (errno == EPERM)
-			ntfs_log_error("%s", hibernated_volume_msg);
-		else if (errno == EOPNOTSUPP) {
+			break;
+		case NTFS_VOLUME_HIBERNATED:
+			ntfs_log_error(hibernated_volume_msg, volume, mntpoint);
+			break;
+		case NTFS_VOLUME_UNCLEAN_UNMOUNT:
 			ntfs_log_error(unclean_journal_msg);
-			ntfs_log_error(forced_mount_msg, volume, mntpoint,
+			ntfs_log_error(forced_mount_msg, volume, mntpoint, 
 				       volume, mntpoint);
-		} else if (errno == EBUSY)
+			break;
+		case NTFS_VOLUME_LOCKED:
 			ntfs_log_error("%s", opened_volume_msg);
-		else if (errno == ENXIO)
+			break;
+		case NTFS_VOLUME_RAID:
 			ntfs_log_error("%s", fakeraid_msg);
-		
-		return NULL;
+			break;
+		case NTFS_VOLUME_NO_PRIVILEGE:
+			ntfs_log_error(access_denied_msg, volume);
+			break;
 	}
-
-	return vol;
 }
 
